/*!
 * \file CalculPropagation.h
 * \brief General Calculations for noise propagation with the NMPB 2008 method
 * \author CSTB
 * \version 1.0
 */

#ifndef CALCUL_PROPAGATION_H
#define CALCUL_PROPAGATION_H

#include "PathStructures.h"
#include <vector>
#include <math.h>
#include <map>
#include <stdlib.h>
#include <string.h>
using namespace std;

// on Unix-like systems there is no way_ to hide local classes ant to prevent them from 
// being replaced by other classes with similar names in other libraries. The use of 
// unusual names may help preventing name clashes.

#ifndef _WIN32
#define  Attenuation   _Local_PROPAN8_Attenuation_
#endif

namespace CalculPropagationNMPB
{
	/*!
	* \brief Calculates sound levels sum
	*
	* Used for  5.2.4 Formula (8) and  5.2.5 Formula (9)
	* 
	* \param n the sound levels number
	* 
	* \param levels The sound levels to sum
	* 
	* \returns the sum
	*/
	double SumLevels(int n, double const* levels);

	/*!
	* \brief Calculates long-term sound level for the given the sound levels (homogeneous and favorable) for a path
	*
	* p.18 -  5.2.3 Formula (7)
	* 
	* \param soundLevel_h The sound level in homogeneous conditions
	* 
	* \param soundLevel_f The sound level in downward-refraction conditions
	* 
	* \param favourableProbability The average occurrence of downward-refraction conditions in the direction of the path
	* 
	* \returns The calculated sound level
	* 
	* \throws ERRProbability
	*/
	double SoundLevelForPath(double soundLevel_h, double soundLevel_f, double favourableProbability);

	/*!
	* \brief Calculates the favorable conditions probability for the (SR) direction
	* 
	* \param sourcePos The source position 3D
	* 
	* \param receiverPos The receiver position 3D
	* 
	* \param nbAngles The number of angle probabilities (should be 18)
	* 
	* \param fcpAngles The favorable conditions probabilities for the angles 20, 40, 60, ..., 360
	* 
	* \param angleNorth The north direction (Ox,ON)
	* 
	* \returns the favorable conditions probability for the (SR) direction
	* 
	* \throws ERRAngle
	*/
	double GetFavorableConditionProbability(Position3D const* sourcePos, Position3D const* receiverPos, int nbAngles, double const* fcpAngles, double angleNorth);

	/*!
	* \brief Calculates the Long-term sound level Leq due to one source at point R, in each given frequency band.
	* 
	* \param nbFreq The frequencies number (user data)
	* 
	* \param Lw The sound power levels of the source, in each frequency band (user data)
	* 
	* \param attH Attenuations due to the propagation between source and receiver in homogeneous conditions, in each frequency band (user data)
	* 
	* \param attF Attenuations due to the propagation between source and receiver in downward-refraction conditions, in each frequency band (user data)
	* 
	* \param fcp Probability of occurrence of downward-refraction conditions over a long-term period in a given direction, p in [0, 1] (user data)
	* 
	* \param LeqH Sound levels due to source Si at point R in homogeneous conditions, in each frequency band (calculated in this function)
	* 
	* \param LeqF Sound levels due to source Si at point R in downward-refraction conditions, in each frequency band (calculated in this function)
	* 
	* \param LeqLT Long-term sound levels due to source Si at point R, in each given frequency band (calculated in this function)
	*/
	void CalculateLeqLT(int nbFreq, double const* Lw, double const* attH, double const* attF, double fcp, double* LeqH, double* LeqF, double* LeqLT);

    /*!
	 * \brief Class used to sound attenuation calculations along a given path, with a specified frequency
	 */
	class Attenuation
    {
    public:

		/*!
		* \brief Initialization of Attenuation class with the path and the frequency 
		* 
		* \param path The path where the attenuation will be calculated
		* 
		* \param frequency The frequency to calculation sound attenuation
		*/
		Attenuation(PropagationPath* path, double frequency);

		/*!
		* \brief Initialization of Attenuation class with the path, the frequency and the favorable probability
		* 
		* \param path The path where the attenuation will be calculated
		* 
		* \param frequency The frequency to calculation sound attenuation
		* 
		* \param favourableProbability The favorable probability
		*/
		Attenuation(PropagationPath* path, double frequency, double favourableProbability);

		/*!
		* \brief Attenuation calculation following geometric divergence
		*
		* Method for attenuation calculation following geometric divergence
		*   9.1, formula (20), p. 37
		* 
		* \param dist : Direct distance between source and receiver.
		* 
		* \returns the calculated attenuation.
		*/
		double DivergenceAttenuationCalculation(double dist);

		/*!
		* \brief Atmospheric absorption
		* 
		* Method to calculate atmospheric absorption attenuation 
		*  9.2, formula (21), p. 37
		*
		* \param dist : direct distance between source and receptor
		* 
		* \param freq : third octave bands center frequency
		* 
		* \returns the calculated absorption attenuation.
		*/
		double AtmosphericAbsorptionCalculation(double dist, double freq);

		/*!
		* \brief Calculates border attenuation, in homogeneous and favorable conditions
		* 
		* Method to calculate border attenuation, in homogeneous and favorable conditions
		*  5.2.1 and 5.2.2, p.17-18
		* 
		* \param nbSideDiffractions The number of side diffractions in the path
		* 
		* \throws ERRNoPoint, ERROnePoint, ERRAttCoeffFrequency
		*
		*/
		void BoundaryAttenuationCalculation(int nbSideDiffractions);

		/*!
		* \brief Calculates total attenuation along the propagation path, in homogeneous and favorable conditions
		* 
		* Method to calculate total attenuation along the propagation path, in homogeneous and favorable conditions
		*  5.2.1, formula (4), p.17 and  5.2.2, formula (6), p.17
		* 
		* \throws ERRSideDiff
		*/
        void AttenuationCalculation();

		/*!
		* \brief Fills the _attenuationCoeffByFrequencyMap map with values of another map
		* 
		* \param newMap The map to get values
		*/
		void FillFrequencyMap(map<int,double> newMap);

		/*!
		 * \brief Gets attenuation along the path in homogeneous conditions, for the given frequency
		 * 
		 * \returns attenuation along the path in homogeneous conditions, for the given frequency
		 */
		double getHomogeneousAttenuation()
		{
			return _attenuation_h;
		}

		/*!
		* \brief Gets attenuation along the path in favorable conditions, for the given frequency
		* 
		* \returns attenuation along the path in favorable conditions, for the given frequency
		*/
		double getFavorableAttenuation()
		{
			return _attenuation_f;
		}

		/*!
		* \brief Gets divergence attenuation
		* 
		* \returns divergence attenuation
		*/
		double getDivergenceAttenuation()
		{
			return _divergenceAttenuation;
		}

		/*!
		* \brief Gets atmospheric absorption
		* 
		* \returns atmospheric absorption
		*/
		double getAtmosphericAbsorption()
		{
			return _atmosphericAbsorption;
		}

		/*!
		* \brief Gets ground attenuation in homogeneous conditions
		* 
		* \returns ground attenuation in homogeneous conditions
		*/
		double getGroundAttenuation_h()
		{
			return _groundAttenuation_h;
		}

		/*!
		* \brief Gets ground attenuation in favorable conditions
		* 
		* \returns ground attenuation in favorable conditions
		*/
		double getGroundAttenuation_f()
		{
			return _groundAttenuation_f;
		}
		
		/*!
		* \brief Gets diffraction attenuation in homogeneous conditions
		* 
		* \returns diffraction attenuation in homogeneous conditions
		*/
		double getDiffractionAttenuation_h()
		{
			return _diffractionAttenuation_h;
		}
		
		/*!
		* \brief Gets diffraction attenuation in favorable conditions
		* 
		* \returns diffraction attenuation in favorable conditions
		*/
		double getDiffractionAttenuation_f()
		{
			return _diffractionAttenuation_f;
		}

		/*!
		* \brief Gets embankment attenuation
		* 
		* \returns embankment attenuation
		*/
		double getEmbankmentAttenuation()
		{
			return _embankmentAttenuation_h;
		}

		/*!
		* \brief Gets border attenuation in homogeneous conditions
		* 
		* \returns border attenuation in homogeneous conditions
		*/
		double getBoundaryAttenuation_h()
		{
			return _boundaryAttenuation_h;
		}

		/*!
		* \brief Gets border attenuation in favorable conditions
		* 
		* \returns border attenuation in favorable conditions
		*/
		double getBoundaryAttenuation_f()
		{
			return _boundaryAttenuation_f;
		}

	private:
		/*!
		 * \brief the propagation path
		 */
		PropagationPath* _path; 
		/*!
		* \brief the frequency
		*/
		double _frequency; 
		/*!
		* \brief the favorable probability
		*/
		double _favourableProbability;

		/*!
		* \brief attenuation along the path in homogeneous conditions
		*/
		double _attenuation_h;
		/*!
		* \brief attenuation along the path in favorable conditions
		*/
		double _attenuation_f;
		/*!
		* \brief divergence attenuation along the path
		*/
		double _divergenceAttenuation;
		/*!
		* \brief atmospheric absorption along the path
		*/
		double _atmosphericAbsorption;
		/*!
		* \brief border attenuation in homogeneous conditions
		*/
		double _boundaryAttenuation_h;
		/*!
		* \brief border attenuation in favorable conditions
		*/
		double _boundaryAttenuation_f;
		/*!
		* \brief ground attenuation in homogeneous conditions (used in border attenuation calculation)
		*/
		double _groundAttenuation_h;
		/*!
		* \brief ground attenuation in favorable conditions (used in border attenuation calculation)
		*/
		double _groundAttenuation_f;
		/*!
		* \brief diffraction attenuation in homogeneous conditions (used in border attenuation calculation)
		*/
		double _diffractionAttenuation_h;
		/*!
		* \brief diffraction attenuation in favorable conditions (used in border attenuation calculation)
		*/
		double _diffractionAttenuation_f;
		/*!
		* \brief embankment attenuation in homogeneous conditions (used in border attenuation calculation)
		*/
		double _embankmentAttenuation_h;
		/*!
		* \brief embankment attenuation in favorable conditions (used in border attenuation calculation)
		*/
		double _embankmentAttenuation_f;

		/*!
		* \brief The atmospheric attenuation coefficients by frequency map
		*/
		map<int,double> _attenuationCoeffByFrequencyMap;

		/*!
		* \brief Fills the _attenuationCoeffByFrequencyMap map with values of the third octave bands
		*
		* Values taken in the Array 2, p. 37 
		*/
		void FillFrequencyMap();

		// to know if calculation steps must be displayed on console
		bool _displayConsoleStep;

		// enable/disable correction for low barrier heights 
		bool _withCh ;
		/*!
		* \brief Calculates ground attenuation, in homogeneous and favorable conditions
		* 
		* Method to calculate ground attenuation, in homogeneous and favorable conditions
		*  9.3 p.38-44
		*
		*/
		void GroundAttenuationCalculation();

		/*!
		* \brief Calculates embankment attenuation
		*
		*  9.3.5 p.43-44
		*
		*/
		void EmbankmentAttenuationCalculation();

		/*!
		* \brief Gets attenuation coefficient by frequency
		*
		* Gets atmospheric attenuation coefficient for the given frequency
		* 
		* \param freq : Nominal median frequency
		* 
		* \returns the attenuation coefficient
		* 
		* \throws ERRAttCoeffFrequency
		*/
		double AttenuationCoeffByFrequency(double freq);

		/*!
		* \brief Clear calculations results
		*/
		void ClearResults();

    };
}

#endif